/*
 * Transcoder.Managed Transcoder_clr.h
 * Copyright (C) 2007-2008 doubleTwist Corporation
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "Transcode.h"
#include "ProgressAdapter.h"
#include <vcclr.h>

using namespace System;
using namespace System::IO;
using namespace System::Runtime::InteropServices;
using namespace System::Diagnostics;
using namespace System::Collections::Generic;
using namespace Dune::Components::Common;

namespace Dune {
namespace Components {
namespace Transcode {

	typedef int (*ReadCallback)(unsigned char *buf, int size);
	typedef int (*WriteCallback)(unsigned char *buf, int size);
	typedef __int64 (*SeekCallback)(__int64 pos, int whence);
	typedef int (*CloseCallback)();

	class StreamAdapter : public transcode::Stream
	{
	public:
		StreamAdapter(ReadCallback readCallback, WriteCallback writeCallback, SeekCallback seekCallback, CloseCallback closeCallback)
		 : mReadCallback(readCallback)
		 , mWriteCallback(writeCallback)
		 , mSeekCallback(seekCallback)
		 , mCloseCallback(closeCallback)
		{
		}

		virtual int Read(unsigned char *buf, int size)
		{
			return mReadCallback(buf, size);
		}

		virtual int Write(unsigned char *buf, int size)
		{
			return mWriteCallback(buf, size);
		}

		virtual __int64 Seek(__int64 pos, int whence)
		{
			return mSeekCallback(pos, whence);
		}

		virtual int Close()
		{
			return mCloseCallback();
		}

	private:
		ReadCallback mReadCallback;
		WriteCallback mWriteCallback;
		SeekCallback mSeekCallback;
		CloseCallback mCloseCallback;
	};

	ref class StreamAdapterClr
	{
		public: delegate int ReadDelegate(unsigned char* buf, int size);
		public: delegate int WriteDelegate(unsigned char* buf, int size);
		public: delegate __int64 SeekDelegate(__int64 pos, int whence);
		public: delegate int CloseDelegate();

		public: StreamAdapterClr(Stream^ stream)
		 : mStream(stream)
		{
		}

		public: int Read(unsigned char *buf, int size)
		{
			array<Byte>^ buffer = gcnew array<Byte>(size);
			int bytesRead = mStream->Read(buffer, 0, size);

			for (int i = 0; i < bytesRead; i++)
				buf[i] = buffer[i];
			
			return bytesRead;
		}

		int Write(unsigned char *buf, int size)
		{
			array<Byte>^ buffer = gcnew array<Byte>(size);

			for (int i = 0; i < size; i++)
				buffer[i] = buf[i];

			mStream->Write(buffer, 0, size);

			return size;
		}

		__int64 Seek(__int64 pos, int whence)
		{
			__int64 newPos = -1;
			try
			{
				switch (whence)
				{
					case SEEK_SET:
						newPos = mStream->Seek(pos, SeekOrigin::Begin);
						break;

					case SEEK_CUR:
						newPos = mStream->Seek(pos, SeekOrigin::Current);
						break;

					case SEEK_END:
						newPos = mStream->Seek(pos, SeekOrigin::End);
						break;

					// Special case for ffmpeg: this 'whence' parameter indicates that ffmpeg wants to know the size of the stream.
					case 0x10000:
						newPos = mStream->Length;
						break;
				}
			}
			catch(...)
			{
				newPos = -1;
			}

			return newPos;
		}

		int Close()
		{
			//mStream->Close();
			return 0;
		}

		private: Stream^ mStream;
	};

	public ref class RemoteTranscoder : public MarshalByRefObject, public ITranscoder
	{
		static Dictionary<int, MediaType>^ sMediaTypeMap = gcnew Dictionary<int, MediaType>();
		static Dictionary<int, FileFormat>^ sFileFormatMap = gcnew Dictionary<int, FileFormat>();
		static Dictionary<int, VideoEncoding>^ sVideoEncodingMap = gcnew Dictionary<int, VideoEncoding>();
		static Dictionary<int, AudioEncoding>^ sAudioEncodingMap = gcnew Dictionary<int, AudioEncoding>();

		static RemoteTranscoder()	// type constructor
		{
			sMediaTypeMap[transcode::eMediaType_Unknown] = MediaType::Unknown;
			sMediaTypeMap[transcode::eMediaType_Audio] = MediaType::Audio;
			sMediaTypeMap[transcode::eMediaType_Video] = MediaType::Video;

			sFileFormatMap[transcode::eFileFormat_Unknown] = FileFormat::Unknown;
			sFileFormatMap[transcode::eFileFormat_MP4] = FileFormat::MP4;
			sFileFormatMap[transcode::eFileFormat_ASF] = FileFormat::ASF;
			sFileFormatMap[transcode::eFileFormat_AVI] = FileFormat::AVI;
			sFileFormatMap[transcode::eFileFormat_FLV] = FileFormat::FLV;
			sFileFormatMap[transcode::eFileFormat_MPEG] = FileFormat::MPEG;
			sFileFormatMap[transcode::eFileFormat_WAV] = FileFormat::WAV;

			sVideoEncodingMap[transcode::eVideoEncoding_Unknown] = VideoEncoding::Unknown;
			sVideoEncodingMap[transcode::eVideoEncoding_MP4] = VideoEncoding::MP4;
			sVideoEncodingMap[transcode::eVideoEncoding_H263] = VideoEncoding::H263;
			sVideoEncodingMap[transcode::eVideoEncoding_H264] = VideoEncoding::H264;
			sVideoEncodingMap[transcode::eVideoEncoding_WMV7] = VideoEncoding::WMV7;
			sVideoEncodingMap[transcode::eVideoEncoding_WMV8] = VideoEncoding::WMV8;
			sVideoEncodingMap[transcode::eVideoEncoding_WMV9] = VideoEncoding::WMV9;
			sVideoEncodingMap[transcode::eVideoEncoding_VC1] = VideoEncoding::VC1;
			sVideoEncodingMap[transcode::eVideoEncoding_DVVideo] = VideoEncoding::DVVideo;
			sVideoEncodingMap[transcode::eVideoEncoding_MPEG1] = VideoEncoding::MPEG1;
			sVideoEncodingMap[transcode::eVideoEncoding_MPEG2] = VideoEncoding::MPEG2;
			sVideoEncodingMap[transcode::eVideoEncoding_H263Spark] = VideoEncoding::H263Spark;
			sVideoEncodingMap[transcode::eVideoEncoding_VP6] = VideoEncoding::VP6;

			sAudioEncodingMap[transcode::eAudioEncoding_Unknown] = AudioEncoding::Unknown;
			sAudioEncodingMap[transcode::eAudioEncoding_AMR_NB] = AudioEncoding::AMR_NB;
			sAudioEncodingMap[transcode::eAudioEncoding_AMR_WB] = AudioEncoding::AMR_WB;
			sAudioEncodingMap[transcode::eAudioEncoding_AAC] = AudioEncoding::AAC;
			sAudioEncodingMap[transcode::eAudioEncoding_AC3] = AudioEncoding::AC3;
			sAudioEncodingMap[transcode::eAudioEncoding_MP3] = AudioEncoding::MP3;
			sAudioEncodingMap[transcode::eAudioEncoding_WMA1] = AudioEncoding::WMA1;
			sAudioEncodingMap[transcode::eAudioEncoding_WMA2] = AudioEncoding::WMA2;
			sAudioEncodingMap[transcode::eAudioEncoding_DVAudio] = AudioEncoding::DVAudio;
			sAudioEncodingMap[transcode::eAudioEncoding_PCM] = AudioEncoding::PCM;
		}

		public: RemoteTranscoder()
		{
		}

		~RemoteTranscoder()
		{
		}

		public: virtual void Transcode(array<String^>^ transcodeArgs, IntPtr progress)
		{
			Transcode(nullptr, transcodeArgs, progress);
		}

		public: virtual void Transcode(array<Stream^>^ inputStreams, array<String^>^ transcodeArgs, IntPtr progress)
		{
			std::vector<transcode::Stream*> inputStreamVector;
			List<GCHandle>^ globalHandles = gcnew List<GCHandle>();

			try
			{
				if (inputStreams != nullptr)
				{
					for (int i = 0; i < inputStreams->Length; i++)
						inputStreamVector.push_back(CreateStreamAdapter(inputStreams[i], globalHandles));
				}

				std::vector<std::string> args;
				for (int i = 0; i < transcodeArgs->Length; i++)
				{
					IntPtr tmpStr = Marshal::StringToHGlobalAnsi(transcodeArgs[i]);
					args.push_back((char*)(void*)tmpStr);
					Marshal::FreeHGlobal(tmpStr);
				}

				IProgress^ remoteProgress = dynamic_cast<IProgress^ >(Activator::GetObject(IProgress::typeid, "ipc://DTApplicationPort/Progress.rem"));
				remoteProgress->SetCallback(progress);

				ProgressDelegate^ fp = gcnew ProgressDelegate(remoteProgress, &IProgress::Update);
				GCHandle delegateGH = GCHandle::Alloc(fp);
				IntPtr ip = Marshal::GetFunctionPointerForDelegate(fp);
				ProgressCallback cb = static_cast<ProgressCallback>(ip.ToPointer());

				ProgressAdapter progressAdapter(cb);

				try
				{
					transcode::Transcode(inputStreamVector, args, progressAdapter);
				}
				finally
				{
					remoteProgress->SetCallback(IntPtr::Zero);
					delegateGH.Free();
				}
			}
			finally
			{
				for (std::vector<transcode::Stream*>::iterator streamIter = inputStreamVector.begin(); streamIter != inputStreamVector.end(); streamIter++)
					delete *streamIter;

				for (int i = 0; i < globalHandles->Count; i++)
					globalHandles[i].Free();
			}
		}

		public: virtual IMetadata^ GetMetadata(Stream^ inputStream, String^ fileExtension)
		{
			List<GCHandle>^ globalHandles = gcnew List<GCHandle>();

			try
			{
				std::auto_ptr<StreamAdapter> streamAdapter(CreateStreamAdapter(inputStream, globalHandles));
				transcode::Metadata nativeMetadata = transcode::GetMetadata(streamAdapter.get());
				Metadata^ metadata = gcnew Metadata();

				// General metadata
				metadata->Type = sMediaTypeMap[nativeMetadata.Type()];
				metadata->FileFormat = sFileFormatMap[nativeMetadata.Format()];
				metadata->HasAudio = nativeMetadata.HasAudio();
				metadata->Duration = nativeMetadata.Duration();
				metadata->AverageBitrate = nativeMetadata.AverageBitrate();
				metadata->Title = Marshal::PtrToStringAnsi((IntPtr)(void*)nativeMetadata.Title().c_str());
				metadata->Artist = Marshal::PtrToStringAnsi((IntPtr)(void*)nativeMetadata.Artist().c_str());
				metadata->Album = Marshal::PtrToStringAnsi((IntPtr)(void*)nativeMetadata.Album().c_str());
				metadata->Genre = Marshal::PtrToStringAnsi((IntPtr)(void*)nativeMetadata.Genre().c_str());

				// Video metadata
				transcode::VideoMetadata& nativeVideoMetadata = nativeMetadata.VideoMetadata();
				IVideoMetadata^ videoMetadata = metadata->VideoMetadata;
				videoMetadata->Duration = nativeVideoMetadata.Duration();
				videoMetadata->Width = nativeVideoMetadata.Width();
				videoMetadata->Height = nativeVideoMetadata.Height();
				videoMetadata->AverageBitrate = nativeVideoMetadata.AverageBitrate();
				videoMetadata->FrameRate = nativeVideoMetadata.FrameRate();
				videoMetadata->PixelAspectRatio = nativeVideoMetadata.PixelAspectRatio();
				videoMetadata->Encoding = sVideoEncodingMap[nativeVideoMetadata.Encoding()];
				videoMetadata->Profile = nativeVideoMetadata.Profile();
				videoMetadata->Level = nativeVideoMetadata.Level();

				// Audio metadata
				transcode::AudioMetadata& nativeAudioMetadata = nativeMetadata.AudioMetadata();
				IAudioMetadata^ audioMetadata = metadata->AudioMetadata;
				audioMetadata->Duration = nativeAudioMetadata.Duration();
				audioMetadata->Bitrate = nativeAudioMetadata.Bitrate();
				audioMetadata->SampleSize = nativeAudioMetadata.SampleSize();
				audioMetadata->SampleRate = nativeAudioMetadata.SampleRate();
				audioMetadata->Encoding = sAudioEncodingMap[nativeAudioMetadata.Encoding()];

				if (nativeAudioMetadata.ChannelCount() == 1)
					audioMetadata->Channels = Channels::Mono;
				else if (nativeAudioMetadata.ChannelCount() == 2)
					audioMetadata->Channels = Channels::Stereo;
				else if (nativeAudioMetadata.ChannelCount() > 2)
					audioMetadata->Channels = Channels::Surround;
				

				// Hack to recognize the 3GP/3G2 file formats.  The transcoder can't tell the difference
				// between MP4 and 3GP.
				if (fileExtension == ".3gp")
					metadata->FileFormat = FileFormat::ThreeGP;
				else if (fileExtension == ".3g2")
					metadata->FileFormat = FileFormat::ThreeG2;

				return metadata;
			}
			finally
			{
				for (int i = 0; i < globalHandles->Count; i++)
					globalHandles[i].Free();
			}
		}

		StreamAdapter* CreateStreamAdapter(Stream^ inputStream, List<GCHandle>^ outGlobalHandles)
		{
			StreamAdapterClr^ streamAdapterClr = gcnew StreamAdapterClr(inputStream);

			GCHandle delegateGH = GCHandle::Alloc(gcnew StreamAdapterClr::ReadDelegate(streamAdapterClr, &StreamAdapterClr::Read));
			outGlobalHandles->Add(delegateGH);
			ReadCallback readCallback = static_cast<ReadCallback>(Marshal::GetFunctionPointerForDelegate((Delegate^)delegateGH.Target).ToPointer());

			delegateGH = GCHandle::Alloc(gcnew StreamAdapterClr::WriteDelegate(streamAdapterClr, &StreamAdapterClr::Write));
			outGlobalHandles->Add(delegateGH);
			WriteCallback writeCallback = static_cast<WriteCallback>(Marshal::GetFunctionPointerForDelegate((Delegate^)delegateGH.Target).ToPointer());

			delegateGH = GCHandle::Alloc(gcnew StreamAdapterClr::SeekDelegate(streamAdapterClr, &StreamAdapterClr::Seek));
			outGlobalHandles->Add(delegateGH);
			SeekCallback seekCallback = static_cast<SeekCallback>(Marshal::GetFunctionPointerForDelegate((Delegate^)delegateGH.Target).ToPointer());

			delegateGH = GCHandle::Alloc(gcnew StreamAdapterClr::CloseDelegate(streamAdapterClr, &StreamAdapterClr::Close));
			outGlobalHandles->Add(delegateGH);
			CloseCallback closeCallback = static_cast<CloseCallback>(Marshal::GetFunctionPointerForDelegate((Delegate^)delegateGH.Target).ToPointer());

			return new StreamAdapter(readCallback, writeCallback, seekCallback, closeCallback);
		}

		public: virtual void SetLogFile(String^ filePath)
		{
			IntPtr tmpStr = Marshal::StringToHGlobalAnsi(filePath);

			try
			{
				transcode::SetLogFile((char*)(void*)tmpStr);
			}
			finally
			{
				Marshal::FreeHGlobal(tmpStr);
			}
		}
	};

}
}
}
