/*
 * Transcoder.Native Metadata.cpp
 * Copyright (C) 2007-2008 doubleTwist Corporation
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Transcode.h"
#include "LogFileOpener.h"

#pragma warning(push)
#pragma warning(disable : 4244)
extern "C"
{
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <sys/stat.h>
#include "avformat.h"
//#include "swscale.h"
//#include "framehook.h"
#include "opt.h"
#include "fifo.h"
#include "avstring.h"
#include "float_cast.h"
#include "DtStream.h"
}
#pragma warning(pop)

#define INT64_C(x) x##LL

using transcode::Metadata;
using namespace transcode;

int DtStreamRead(void* userData, unsigned char *buf, int size);
int DtStreamWrite(void* userData, unsigned char *buf, int size);
offset_t DtStreamSeek(void* userData, offset_t pos, int whence);
int DtStreamClose(void* userData);

struct FormatMapPair
{
	std::string mName;
	FileFormat mType;
};

static const int kFormatMapSize = 7;
static const FormatMapPair sFormatMap[kFormatMapSize] =
{
	{ "mov,mp4,m4a,3gp,3g2,mj2", eFileFormat_MP4 },
	{ "asf", eFileFormat_ASF },
	{ "avi", eFileFormat_AVI },
	{ "flv", eFileFormat_FLV },
	{ "mpeg", eFileFormat_MPEG },
	{ "mp3", eFileFormat_MPEG },
	{ "wav", eFileFormat_WAV }
};

struct VideoEncodingMapPair
{
	CodecID mID;
	VideoEncoding mType;
};

struct AudioEncodingMapPair
{
	CodecID mID;
	AudioEncoding mType;
};

static const int kVideoEncodingMapSize = 12;
static const VideoEncodingMapPair sVideoEncodingMap[kVideoEncodingMapSize] =
{
	{ CODEC_ID_MPEG4, eVideoEncoding_MP4 },
	{ CODEC_ID_H263, eVideoEncoding_H263 },
	{ CODEC_ID_H264, eVideoEncoding_H264 },
	{ CODEC_ID_WMV1, eVideoEncoding_WMV7 },
	{ CODEC_ID_WMV2, eVideoEncoding_WMV8 },
	{ CODEC_ID_WMV3, eVideoEncoding_WMV9 },
	{ CODEC_ID_VC1, eVideoEncoding_VC1 },
	{ CODEC_ID_DVVIDEO, eVideoEncoding_DVVideo },
	{ CODEC_ID_MPEG1VIDEO, eVideoEncoding_MPEG1 },
	{ CODEC_ID_MPEG2VIDEO, eVideoEncoding_MPEG2 },
	{ CODEC_ID_FLV1, eVideoEncoding_H263Spark },
	{ CODEC_ID_VP6, eVideoEncoding_VP6 }
};

static const int kAudioEncodingMapSize = 9;
static const AudioEncodingMapPair sAudioEncodingMap[kAudioEncodingMapSize] =
{
	{ CODEC_ID_AMR_NB, eAudioEncoding_AMR_NB },
	{ CODEC_ID_AMR_WB, eAudioEncoding_AMR_WB },
	{ CODEC_ID_AAC, eAudioEncoding_AAC },
	{ CODEC_ID_AC3, eAudioEncoding_AC3 },
	{ CODEC_ID_MP3, eAudioEncoding_MP3 },
	{ CODEC_ID_WMAV1, eAudioEncoding_WMA1 },
	{ CODEC_ID_WMAV2, eAudioEncoding_WMA2 },
	{ CODEC_ID_DVAUDIO, eAudioEncoding_DVAudio },
	{ CODEC_ID_PCM_S16LE, eAudioEncoding_PCM }
};

static FileFormat AsFormat(const char* name)
{
	FileFormat value = eFileFormat_Unknown;

	for (int i = 0; i < kFormatMapSize; ++i)
	{
		if (sFormatMap[i].mName == name)
		{
			value = sFormatMap[i].mType;
			break;
		}
	}

	return value;
}

static VideoEncoding AsVideoEncoding(CodecID id)
{
	VideoEncoding value = eVideoEncoding_Unknown;

	for (int i = 0; i < kVideoEncodingMapSize; ++i)
	{
		if (sVideoEncodingMap[i].mID == id)
		{
			value = sVideoEncodingMap[i].mType;
			break;
		}
	}

	return value;
}

static AudioEncoding AsAudioEncoding(CodecID id)
{
	AudioEncoding value = eAudioEncoding_Unknown;

	if ((id & 0xff000) == CODEC_ID_PCM_S16LE)
	{
		// Special case for PCM audio because we map a bunch of specific types into one.
		value = eAudioEncoding_PCM;
	}
	else
	{
		for (int i = 0; i < kAudioEncodingMapSize; ++i)
		{
			if (sAudioEncodingMap[i].mID == id)
			{
				value = sAudioEncodingMap[i].mType;
				break;
			}
		}
	}

	return value;
}

VideoMetadata::VideoMetadata()
 : mDuration(0.0f)
 , mAverageBitrate(0)
 , mWidth(0)
 , mHeight(0)
 , mFrameRate(0.0f)
 , mPixelAspectRatio(1.0f)
 , mEncoding(eVideoEncoding_Unknown)
 , mProfile(0)
 , mLevel(0)
{
}

VideoMetadata::~VideoMetadata()
{
}

int VideoMetadata::Width() const
{
	return mWidth;
}

void VideoMetadata::Width(int value)
{
	mWidth = value;
}

int VideoMetadata::Height() const
{
	return mHeight;
}

void VideoMetadata::Height(int value)
{
	mHeight = value;
}

long VideoMetadata::AverageBitrate() const
{
	return mAverageBitrate;
}

void VideoMetadata::AverageBitrate(long value)
{
	mAverageBitrate = value;
}

float VideoMetadata::Duration() const
{
	return mDuration;
}

void VideoMetadata::Duration(float value)
{
	mDuration = value;
}

float VideoMetadata::FrameRate() const
{
	return mFrameRate;
}

void VideoMetadata::FrameRate(float value)
{
	mFrameRate = value;
}

float VideoMetadata::PixelAspectRatio() const
{
	return mPixelAspectRatio;
}

void VideoMetadata::PixelAspectRatio(float value)
{
	mPixelAspectRatio = value;
}

VideoEncoding VideoMetadata::Encoding() const
{
	return mEncoding;
}

void VideoMetadata::Encoding(VideoEncoding value)
{
	mEncoding = value;
}

int VideoMetadata::Profile() const
{
	return mProfile;
}

void VideoMetadata::Profile(int value)
{
	mProfile = value;
}

int VideoMetadata::Level() const
{
	return mLevel;
}

void VideoMetadata::Level(int value)
{
	mLevel = value;
}

AudioMetadata::AudioMetadata()
 : mDuration(0.0f)
 , mBitrate(0)
 , mSampleSize(0)
 , mSampleRate(0)
 , mChannelCount(0)
 , mEncoding(eAudioEncoding_Unknown)
{
}

AudioMetadata::~AudioMetadata()
{
}

void AudioMetadata::Encoding(AudioEncoding value)
{
	mEncoding = value;
}

AudioEncoding AudioMetadata::Encoding() const
{
	return mEncoding;
}

float AudioMetadata::Duration() const
{
	return mDuration;
}

void AudioMetadata::Duration(float value)
{
	mDuration = value;
}

int AudioMetadata::Bitrate() const
{
	return mBitrate;
}

void AudioMetadata::Bitrate(int value)
{
	mBitrate = value;
}

int AudioMetadata::SampleSize() const
{
	return mSampleSize;
}

void AudioMetadata::SampleSize(int value)
{
	mSampleSize = value;
}

int AudioMetadata::SampleRate() const
{
	return mSampleRate;
}

void AudioMetadata::SampleRate(int value)
{
	mSampleRate = value;
}

int AudioMetadata::ChannelCount() const
{
	return mChannelCount;
}

void AudioMetadata::ChannelCount(int value)
{
	mChannelCount = value;
}

Metadata::Metadata()
 : mDuration(0.0f)
 , mAverageBitrate(0)
 , mMediaType(eMediaType_Unknown)
 , mFileFormat(eFileFormat_Unknown)
 , mHasAudio(false)
{
}
/*
Metadata::Metadata(const Metadata& rhs)
 : mDuration(rhs.mDuration)
 , mAverageBitrate(rhs.mAverageBitrate)
 , mMediaType(rhs.mMediaType)
 , mFileFormat(rhs.mFileFormat)
 , mHasAudio(rhs.mHasAudio)
 , mVideoMetadata(rhs.mVideoMetadata)
 , mAudioMetadata(rhs.mAudioMetadata)
{
}
*/
Metadata::~Metadata()
{
}

float Metadata::Duration() const
{
	return mDuration;
}

void Metadata::Duration(float value)
{
	mDuration = value;
}

long Metadata::AverageBitrate() const
{
	return mAverageBitrate;
}

void Metadata::AverageBitrate(long value)
{
	mAverageBitrate = value;
}

MediaType Metadata::Type() const
{
	return mMediaType;
}

void Metadata::Type(MediaType value)
{
	mMediaType = value;
}

FileFormat Metadata::Format() const
{
	return mFileFormat;
}

void Metadata::Format(FileFormat value)
{
	mFileFormat = value;
}

bool Metadata::HasAudio() const
{
	return mHasAudio;
}

void Metadata::HasAudio(bool value)
{
	mHasAudio = value;
}

std::string Metadata::Title() const
{
	return mTitle;
}

void Metadata::Title(const std::string& value)
{
	mTitle = value;
}

std::string Metadata::Artist() const
{
	return mArtist;
}

void Metadata::Artist(const std::string& value)
{
	mArtist = value;
}

std::string Metadata::Album() const
{
	return mAlbum;
}

void Metadata::Album(const std::string& value)
{
	mAlbum = value;
}

std::string Metadata::Genre() const
{
	return mGenre;
}

void Metadata::Genre(const std::string& value)
{
	mGenre = value;
}

VideoMetadata& Metadata::VideoMetadata()
{
	return mVideoMetadata;
}

AudioMetadata& Metadata::AudioMetadata()
{
	return mAudioMetadata;
}

Metadata transcode::GetMetadata(Stream* inputStream)
{
    AVFormatContext *pFormatCtx;
    unsigned int    i, videoStream, audioStream;
    Metadata		metadata;

	// This will open and close the log file automatically.
	av_log_set_level(AV_LOG_DEBUG);
	LogFileOpener logOpener;

    // Register all formats and codecs
    av_register_all();

	std::auto_ptr<DtStream> dtStream(new DtStream);
	dtStream->userData = (void*)inputStream;
	dtStream->Read = &DtStreamRead;
	dtStream->Write = &DtStreamWrite;
	dtStream->Seek = &DtStreamSeek;
	dtStream->Close = &DtStreamClose;

	char fakeFileName[512];
	strcpy(fakeFileName, "stream:");
	*(void**)&fakeFileName[strlen(fakeFileName)] = dtStream.get();

    // Open video file
    if(av_open_input_file(&pFormatCtx, fakeFileName, NULL, 0, NULL)!=0)
        throw std::exception("Could not open file");

    // Retrieve stream information
    int result = av_find_stream_info(pFormatCtx);
    if (result < 0)
    {
		pFormatCtx->pb->is_streamed = false;
		result = av_find_stream_info(pFormatCtx);
    }

    // Find the first video stream
    videoStream=-1;
    for(i=0; i<pFormatCtx->nb_streams; i++)
    {
        if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO)
        {
            videoStream=i;
            break;
        }
    }

    // Find the first audio stream.
    audioStream=-1;
    for(i=0; i<pFormatCtx->nb_streams; i++)
    {
        if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_AUDIO)
        {
            audioStream=i;
            break;
        }
    }

	// Set video metadata.
	if (videoStream != -1)
	{
		VideoMetadata& videoMetadata = metadata.VideoMetadata();

		metadata.Type(eMediaType_Video);

		AVStream* stream = pFormatCtx->streams[videoStream];
		AVCodecContext* pCodecCtx = stream->codec;

		// Set duration.
		if (stream->duration != AV_NOPTS_VALUE)
		{
			const float kTimeBase = stream->time_base.num / static_cast<float>(stream->time_base.den);
			videoMetadata.Duration(stream->duration * kTimeBase);
		}

		// Set bitrate.
		videoMetadata.AverageBitrate(pCodecCtx->bit_rate);

		// Set the encoding type.
		videoMetadata.Encoding(AsVideoEncoding(pCodecCtx->codec_id));

		// Set frame width and height.
		videoMetadata.Width(pCodecCtx->width);
		videoMetadata.Height(pCodecCtx->height);

		// Set the frame rate.
		if (stream->r_frame_rate.den > 0)
			videoMetadata.FrameRate(stream->r_frame_rate.num / static_cast<float>(stream->r_frame_rate.den));

		// Set the pixel aspect ratio.
		if (pCodecCtx->sample_aspect_ratio.den > 0)
			videoMetadata.PixelAspectRatio(pCodecCtx->sample_aspect_ratio.num / static_cast<float>(pCodecCtx->sample_aspect_ratio.den));

		videoMetadata.Profile(pCodecCtx->profile);
		videoMetadata.Level(pCodecCtx->level);
	}

	// Set audio metadata.
	if (audioStream != -1)
	{
		AudioMetadata& audioMetadata = metadata.AudioMetadata();

		if (metadata.Type() == eMediaType_Unknown)
			metadata.Type(eMediaType_Audio);

		metadata.HasAudio(true);

		AVStream* stream = pFormatCtx->streams[audioStream];
		AVCodecContext* pCodecCtx = stream->codec;

		// Set the encoding type.
		audioMetadata.Encoding(AsAudioEncoding(pCodecCtx->codec_id));

		// Set audio duration.
		if (stream->duration != AV_NOPTS_VALUE)
		{
			const float kTimeBase = stream->time_base.num / static_cast<float>(stream->time_base.den);
			audioMetadata.Duration(stream->duration * kTimeBase);
		}

		// Set bitrate.
		audioMetadata.Bitrate(pCodecCtx->bit_rate);

		// Set sample size.
		//audioMetadata.SampleSize();		// Comments in ffmpeg code say sample size setting is currently unused.

		// Set sampling rate.
		audioMetadata.SampleRate(pCodecCtx->sample_rate);

		// Set the number of channels.
		audioMetadata.ChannelCount(pCodecCtx->channels);
	}

	// Set overall duration.
	float totalDuration = 0.0f;
    if (pFormatCtx->duration != AV_NOPTS_VALUE)
    {
		totalDuration = pFormatCtx->duration / static_cast<float>(AV_TIME_BASE);
        metadata.Duration(totalDuration);
    }

	// Calculate the average bitrate by dividing total number of bytes in the file by total duration.
	if (totalDuration > 0.0f)
	{
		__int64 streamSize = inputStream->Seek(0, 0x10000);		// HACK: special code passed to Seek() method to get size of stream.  No seeking is actually done.
		metadata.AverageBitrate(static_cast<long>(streamSize * 8 / totalDuration));
	}

	// Set the format type.
	metadata.Format(AsFormat(pFormatCtx->iformat->name));

	// Misc
	metadata.Title(pFormatCtx->title);
	metadata.Artist(pFormatCtx->author);
	metadata.Album(pFormatCtx->album);
	metadata.Genre(pFormatCtx->genre);

	av_close_input_file(pFormatCtx);

    return metadata;
}