/*
 * Various utilities for command line tools
 * Copyright (c) 2000-2003 Fabrice Bellard
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <string.h>
#include <stdlib.h>
#include <errno.h>

class Options;

#pragma warning(push)
#pragma warning(disable : 4244)
extern "C"
{
#include "avformat.h"
#include "avstring.h"
}
#pragma warning(pop)

#include "cmdutils.h"

#undef exit

void show_help_options(const OptionDef *options, const char *msg, int mask, int value)
{
    const OptionDef *po;
    int first;

    first = 1;
    for(po = options; po->name != NULL; po++) {
        char buf[64];
        if ((po->flags & mask) == value) {
            if (first) {
                printf("%s", msg);
                first = 0;
            }
            av_strlcpy(buf, po->name, sizeof(buf));
            if (po->flags & HAS_ARG) {
                av_strlcat(buf, " ", sizeof(buf));
                av_strlcat(buf, po->argname, sizeof(buf));
            }
            printf("-%-17s  %s\n", buf, po->help);
        }
    }
}

static const OptionDef* find_option(const OptionDef *po, const char *name){
    while (po->name != NULL) {
        if (!strcmp(name, po->name))
            break;
        po++;
    }
    return po;
}

void parse_options(const std::vector<std::string>& argv, const OptionDef *optionMap, Options& options)
{
    const char *opt, *arg;
    int handleoptions = 1;
    const OptionDef *po;

	std::vector<std::string>::size_type argc = argv.size();

    /* parse options */
	std::vector<std::string>::size_type optindex = 0;
    while (optindex < argc) {
        opt = argv[optindex++].c_str();

        if (opt[0] == '-' && opt[1] != '\0') {
          if (opt[1] == '-' && opt[2] == '\0') {
            handleoptions = 0;
            continue;
          }
            po= find_option(optionMap, opt + 1);
            if (!po->name)
                po= find_option(optionMap, "default");
            if (!po->name) {
unknown_opt:
                //fprintf(stderr, "%s: unrecognized option '%s'\n", argv[0], opt);
                //exit(1);
                throw std::exception("unrecognized option");
            }
            arg = NULL;
            if (po->flags & HAS_ARG) {
                arg = argv[optindex++].c_str();
                if (!arg) {
                    //fprintf(stderr, "%s: missing argument for option '%s'\n", argv[0], opt);
                    //exit(1);
                    throw std::exception("missing argument for option");
                }
            }
/*
            if (po->flags & OPT_STRING) {
                char *str;
                str = av_strdup(arg);
                *po->u.str_arg = str;
            } else if (po->flags & OPT_BOOL) {
                *po->u.int_arg = 1;
            } else if (po->flags & OPT_INT) {
                *po->u.int_arg = atoi(arg);
            } else if (po->flags & OPT_INT64) {
                *po->u.int64_arg = _strtoi64(arg, (char **)NULL, 10);
            } else if (po->flags & OPT_FLOAT) {
                *po->u.float_arg = atof(arg);
            } else */ if (po->flags & OPT_FUNC2) {
                if(po->u.func2_arg(opt+1, arg, options)<0)
                    goto unknown_opt;
            } else {
                void (*func_arg)(const char *, Options&);
				func_arg = (void (__cdecl *)(const char *, Options&))po->u.func_arg;
				func_arg(arg, options);
            }
        } else {
            parse_arg_file(opt, options);
        }
    }
}

/*
void print_error(const char *filename, int err)
{
    switch(err) {
    case AVERROR_NUMEXPECTED:
        fprintf(stderr, "%s: Incorrect image filename syntax.\n"
                "Use '%%d' to specify the image number:\n"
                "  for img1.jpg, img2.jpg, ..., use 'img%%d.jpg';\n"
                "  for img001.jpg, img002.jpg, ..., use 'img%%03d.jpg'.\n",
                filename);
        break;
    case AVERROR_INVALIDDATA:
        fprintf(stderr, "%s: Error while parsing header\n", filename);
        break;
    case AVERROR_NOFMT:
        fprintf(stderr, "%s: Unknown format\n", filename);
        break;
    case AVERROR(EIO):
        fprintf(stderr, "%s: I/O error occured\n"
                "Usually that means that input file is truncated and/or corrupted.\n",
                filename);
        break;
    case AVERROR(ENOMEM):
        fprintf(stderr, "%s: memory allocation error occured\n", filename);
        break;
    case AVERROR(ENOENT):
        fprintf(stderr, "%s: no such file or directory\n", filename);
        break;
    default:
        fprintf(stderr, "%s: Error while opening file\n", filename);
        break;
    }
}
*/

std::string get_error_string(const char *filename, int err)
{
	std::string retStr(filename);

    switch(err) {
    case AVERROR_NUMEXPECTED:
        retStr += "Incorrect image filename syntax.\n";
        break;
    case AVERROR_INVALIDDATA:
		retStr += "Error while parsing header";
        break;
    case AVERROR_NOFMT:
		retStr += "Unknown format";
        break;
    case AVERROR(EIO):
        retStr += "I/O error occured.  Usually that means that input file is truncated and/or corrupted.";
        break;
    case AVERROR(ENOMEM):
		retStr += "memory allocation error occured";
        break;
    case AVERROR(ENOENT):
        retStr += "%no such file or directory";
        break;
    default:
        retStr += "Error while opening file";
        break;
    }

	return retStr;
}
